=== modified file 'NEWS'
--- NEWS	2006-09-28 16:04:42 +0000
+++ NEWS	2006-09-29 06:24:03 +0000
@@ -9,6 +9,10 @@
       can take a ``working_dir='foo'`` parameter, which will change directory 
       for the command. (John Arbash Meinel)
 
+    * ``Transport.get`` has had its interface made more clear for ease of use.
+      Retrieval of a directory must now fail with either 'PathError' at open
+      time, or raise 'ReadError' on a read. (Robert Collins)
+
   BUG FIXES:
 
     * bzr branch/pull no longer complain about progress bar cleanup when

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-09-27 20:27:46 +0000
+++ bzrlib/errors.py	2006-09-29 06:24:03 +0000
@@ -246,6 +246,10 @@
 
 # XXX: Should be unified with TransportError; they seem to represent the
 # same thing
+# RBC 20060929: I think that unifiying with TransportError would be a mistake
+# - this is finer than a TransportError - and more useful as such. It 
+# differentiates between 'transport has failed' and 'operation on a transport
+# has failed.'
 class PathError(BzrNewError):
     """Generic path error: %(path)r%(extra)s)"""
 
@@ -297,6 +301,10 @@
         PathError.__init__(self, url, extra=extra)
 
 
+class ReadError(PathError):
+    """Error reading from %(path)r."""
+
+
 class ShortReadvError(PathError):
     """readv() read %(actual)s bytes rather than %(length)s bytes at %(offset)s for %(path)s%(extra)s"""
 

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2006-09-27 19:20:35 +0000
+++ bzrlib/tests/test_errors.py	2006-09-29 06:24:03 +0000
@@ -53,6 +53,17 @@
                              repo.bzrdir.root_transport.base,
                              str(error))
 
+    def test_read_error(self):
+        # a unicode path to check that %r is being used.
+        path = u'a path'
+        error = errors.ReadError(path)
+        self.assertEqualDiff("Error reading from u'a path'.", str(error))
+
+    def test_transport_not_possible(self):
+        error = errors.TransportNotPossible('readonly', 'original error')
+        self.assertEqualDiff('Transport operation not possible:'
+                         ' readonly original error', str(error))
+
 
 class PassThroughError(errors.BzrNewError):
     """Pass through %(foo)s and %(bar)s"""
@@ -84,11 +95,3 @@
         e = ErrorWithBadFormat(not_thing='x')
         self.assertStartsWith(
             str(e), 'Unprintable exception ErrorWithBadFormat(')
-
-
-class TestSpecificErrors(TestCase):
-    
-    def test_transport_not_possible(self):
-        e = errors.TransportNotPossible('readonly', 'original error')
-        self.assertEqual('Transport operation not possible:'
-                         ' readonly original error', str(e))

=== modified file 'bzrlib/tests/test_transport.py'
--- bzrlib/tests/test_transport.py	2006-09-13 12:32:47 +0000
+++ bzrlib/tests/test_transport.py	2006-09-29 06:24:03 +0000
@@ -25,6 +25,7 @@
                            TransportNotPossible,
                            ConnectionError,
                            DependencyNotPresent,
+                           ReadError,
                            UnsupportedProtocol,
                            )
 from bzrlib.tests import TestCase, TestCaseInTempDir
@@ -32,6 +33,7 @@
                               _get_protocol_handlers,
                               _get_transport_modules,
                               get_transport,
+                              LateReadError,
                               register_lazy_transport,
                               _set_protocol_handlers,
                               Transport,
@@ -100,6 +102,16 @@
         finally:
             _set_protocol_handlers(saved_handlers)
 
+    def test_LateReadError(self):
+        """The LateReadError helper should raise on read()."""
+        a_file = LateReadError('a path')
+        try:
+            a_file.read()
+        except ReadError, error:
+            self.assertEqual('a path', error.path)
+        self.assertRaises(ReadError, a_file.read, 40)
+        a_file.close()
+
 
 class TestCoalesceOffsets(TestCase):
     

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-09-17 21:24:27 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-09-29 06:24:03 +0000
@@ -122,6 +122,26 @@
         self.assertListRaises(NoSuchFile, t.get_multi, ['a', 'b', 'c'])
         self.assertListRaises(NoSuchFile, t.get_multi, iter(['a', 'b', 'c']))
 
+    def test_get_directory(self):
+        """get() of a directory should only error when read() is called."""
+        t = self.get_transport()
+        if t.is_readonly():
+            self.build_tree(['a directory/'])
+        else:
+            t.mkdir('a%20directory')
+        # getting the file must either work or fail with a PathError
+        try:
+            a_file = t.get('a%20directory')
+        except errors.PathError:
+            # early failure return immediately.
+            return
+        # having got a file, read() must either work (i.e. http reading a dir listing) or
+        # fail with ReadError
+        try:
+            a_file.read()
+        except errors.ReadError:
+            pass
+
     def test_get_bytes(self):
         t = self.get_transport()
 

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2006-09-16 02:26:44 +0000
+++ bzrlib/transport/__init__.py	2006-09-29 06:24:03 +0000
@@ -189,6 +189,23 @@
                    (other.start, other.length, other.ranges))
 
 
+class LateReadError(object):
+    """A helper for transports which pretends to be a readable file.
+
+    When read() is called, errors.ReadError is raised.
+    """
+
+    def __init__(self, path):
+        self._path = path
+
+    def close(self):
+        """a no-op - do nothing."""
+
+    def read(self, count=-1):
+        """Raise ReadError."""
+        raise errors.ReadError(self._path)
+
+
 class Transport(object):
     """This class encapsulates methods for retrieving or putting a file
     from/to a storage location.
@@ -429,6 +446,17 @@
     def get(self, relpath):
         """Get the file at the given relative path.
 
+        This may fail in a number of ways:
+         - HTTP servers may return content for a directory. (unexpected
+           content failure)
+         - FTP servers may indicate NoSuchFile for a directory.
+         - SFTP servers may give a file handle for a directory that will
+           fail on read().
+
+        For correct use of the interface, be sure to catch errors.PathError
+        when calling it and catch errors.ReadError when reading from the
+        returned object.
+
         :param relpath: The relative path to the file
         :rtype: File-like object.
         """

=== modified file 'bzrlib/transport/ftp.py'
--- bzrlib/transport/ftp.py	2006-09-15 01:12:18 +0000
+++ bzrlib/transport/ftp.py	2006-09-29 06:24:03 +0000
@@ -527,7 +527,7 @@
 
 
 class FtpServer(Server):
-    """Common code for SFTP server facilities."""
+    """Common code for FTP server facilities."""
 
     def __init__(self):
         self._root = None

=== modified file 'bzrlib/transport/local.py'
--- bzrlib/transport/local.py	2006-09-06 20:42:45 +0000
+++ bzrlib/transport/local.py	2006-09-29 06:24:03 +0000
@@ -35,7 +35,7 @@
                             check_legal_path, rmtree)
 from bzrlib.symbol_versioning import warn
 from bzrlib.trace import mutter
-from bzrlib.transport import Transport, Server
+from bzrlib.transport import LateReadError, Transport, Server
 
 
 _append_flags = os.O_CREAT | os.O_APPEND | os.O_WRONLY | osutils.O_BINARY
@@ -131,6 +131,8 @@
             path = self._abspath(relpath)
             return open(path, 'rb')
         except (IOError, OSError),e:
+            if e.errno == errno.EISDIR:
+                return LateReadError(relpath)
             self._translate_error(e, path)
 
     def put_file(self, relpath, f, mode=None):

=== modified file 'bzrlib/transport/memory.py'
--- bzrlib/transport/memory.py	2006-09-16 13:36:47 +0000
+++ bzrlib/transport/memory.py	2006-09-29 06:24:03 +0000
@@ -29,7 +29,12 @@
 
 from bzrlib.errors import TransportError, NoSuchFile, FileExists, LockError
 from bzrlib.trace import mutter
-from bzrlib.transport import (Transport, register_transport, Server)
+from bzrlib.transport import (
+    LateReadError,
+    register_transport,
+    Server,
+    Transport,
+    )
 import bzrlib.urlutils as urlutils
 
 
@@ -121,7 +126,10 @@
         """See Transport.get()."""
         _abspath = self._abspath(relpath)
         if not _abspath in self._files:
-            raise NoSuchFile(relpath)
+            if _abspath in self._dirs:
+                return LateReadError(relpath)
+            else:
+                raise NoSuchFile(relpath)
         return StringIO(self._files[_abspath][0])
 
     def put_file(self, relpath, f, mode=None):

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2006-09-18 07:58:41 +0000
+++ bzrlib/transport/sftp.py	2006-09-29 06:24:03 +0000
@@ -50,6 +50,7 @@
 from bzrlib.osutils import pathjoin, fancy_rename, getcwd
 from bzrlib.trace import mutter, warning
 from bzrlib.transport import (
+    LateReadError,
     register_urlparse_netloc_protocol,
     Server,
     split_url,
@@ -392,7 +393,10 @@
                 f.prefetch()
             return f
         except (IOError, paramiko.SSHException), e:
-            self._translate_io_exception(e, path, ': error retrieving')
+            self._translate_io_exception(e, path, ': error retrieving',
+                raise_on_failure=False)
+            # a paramiko generic failure occured.
+            return LateReadError(relpath)
 
     def readv(self, relpath, offsets):
         """See Transport.readv()"""
@@ -682,7 +686,8 @@
         self._mkdir(self._remote_path(relpath), mode=mode)
 
     def _translate_io_exception(self, e, path, more_info='', 
-                                failure_exc=PathError):
+                                failure_exc=PathError,
+                                raise_on_failure=True):
         """Translate a paramiko or IOError into a friendlier exception.
 
         :param e: The original exception
@@ -694,6 +699,8 @@
                            no more information.
                            If this parameter is set, it defines the exception 
                            to raise in these cases.
+        :param raise_on_failure: When False do not raise on a generic paramiko
+            'Failure', rather return to the caller to dispatch differently.
         """
         # paramiko seems to generate detailless errors.
         self._translate_error(e, path, raise_generic=False)
@@ -705,7 +712,10 @@
                 raise FileExists(path, str(e) + more_info)
             # strange but true, for the paramiko server.
             if (e.args == ('Failure',)):
-                raise failure_exc(path, str(e) + more_info)
+                if raise_on_failure:
+                    raise failure_exc(path, str(e) + more_info)
+                else:
+                    return
             mutter('Raising exception with args %s', e.args)
         if getattr(e, 'errno', None) is not None:
             mutter('Raising exception with errno %s', e.errno)

=== modified file 'bzrlib/transport/smart.py'
--- bzrlib/transport/smart.py	2006-09-26 05:51:37 +0000
+++ bzrlib/transport/smart.py	2006-09-29 06:24:03 +0000
@@ -352,7 +352,11 @@
         return SmartServerResponse((r,))
 
     def do_get(self, relpath):
-        backing_bytes = self._backing_transport.get_bytes(relpath)
+        try:
+            backing_bytes = self._backing_transport.get_bytes(relpath)
+        except errors.ReadError:
+            # cannot read the file
+            return SmartServerResponse(('ReadError', ))
         return SmartServerResponse(('ok',), backing_bytes)
 
     def _deserialise_optional_mode(self, mode):
@@ -876,6 +880,12 @@
                 raise UnicodeEncodeError(encoding, val, start, end, reason)
         elif what == "ReadOnlyError":
             raise errors.TransportNotPossible('readonly transport')
+        elif what == "ReadError":
+            if orig_path is not None:
+                error_path = orig_path
+            else:
+                error_path = resp[1]
+            raise errors.ReadError(error_path)
         else:
             raise errors.SmartProtocolError('unexpected smart server error: %r' % (resp,))
 

